package org.netkernel.module.standard.endpoint;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

public abstract class TransparentConfiguredOverlayImpl extends ConfiguredOverlayImpl
{
	public static abstract class TransparentConfig implements IConfig
	{
		private ISpace mDelegateSpace;
		private ISpaceElements mElements;
		private Map<String, IMetaRepresentation> mElementMetasById;
		
		public TransparentConfig(ISpace aDelegateSpace, ISpaceMeta aDelegateMeta, INKFRequestContext aContext, IEndpoint aEndpoint) throws Exception
		{
			mDelegateSpace=aDelegateSpace;
			
			//construct space elements, and get all elements meta
			ISpaceElements delegateElements=aDelegateMeta.getElements();			
			mElementMetasById = new LinkedHashMap<String, IMetaRepresentation>();
			List<String> unreferencedTargets = new ArrayList<String>();
			for (int i=0; i<delegateElements.size(); i++)
			{	IIdentifier id=delegateElements.getIdentifier(i);
				if (id!=null)
				{	String elementId = id.toString();
					unreferencedTargets.add(elementId);
					IMetaRepresentation meta=aContext.meta(elementId);
					int c=0;
					while(meta==null && c++<20) //may be metadata state is not yet consistent - retry usually succeeds
					{	Thread.sleep(100);
						meta=aContext.meta(elementId);
					}
					mElementMetasById.put(elementId, meta);
				}
			}
			// build our list of elements
			TupleList tl=new TupleList(3,delegateElements.size());
			for (int i=0; i<delegateElements.size(); i++)
			{	tl.put(delegateElements.getIdentifier(i),aDelegateSpace,aEndpoint);
			}
			mElements=new SpaceElementsImpl(tl);			
		}
		
		public IMetaRepresentation getElementMeta(String elementId)
		{	return mElementMetasById.get(elementId);
		}

		public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
		{	return mElements;
		}
		
		public ISpace[] getSpaces(INKFRequestContext context) throws Exception
		{	return IStandardEndpoint.NO_SPACES;
		}		

		public String getResolvedElement(INKFResolutionContext aContext) throws Exception
		{	String result=null;
			INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
			for (Map.Entry<String, IMetaRepresentation> entry: mElementMetasById.entrySet())
			{	IMetaRepresentation meta=entry.getValue();
				if (meta instanceof IEndpointMeta)
				{	String id=entry.getKey();
					if (MetadataUtils.match((IEndpointMeta)meta, toResolve, aContext,id))
					{	result = entry.getKey();
						break;
					}
				}
			}
			return result;
		}
		
		protected ISpace getDelegateSpace()
		{	return mDelegateSpace;
		}
		
		public abstract void onRequest(String elementId, INKFRequestContext aContext) throws Exception;
	}
	
	protected IConfig createConfig(INKFRequestContext aContext) throws Exception
	{	return createTransparentConfig(aContext);
	}
	
	protected abstract TransparentConfig createTransparentConfig(INKFRequestContext aContext) throws Exception;
	
	
}